import logging
import os
import argparse
import random
import tqdm
import numpy as np

import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils as utils

from regression_model.script import dataloader, utility, earlystopping
from regression_model.model import models
from datetime import datetime
import matplotlib.pyplot as plt
from regression_model.config.core import *



def set_env(seed):
    # Set available CUDA devices
    # This option is crucial for an multi-GPU device
    os.environ['CUDA_VISIBLE_DEVICES'] = '0, 1'
    # os.environ['CUDA_LAUNCH_BLOCKING'] = '1'
    # os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':16:8'
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
    os.environ['PYTHONHASHSEED']=str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms(True)

def get_parameters():
    parser = argparse.ArgumentParser(description='STGCN')
    parser.add_argument('--enable_cuda', type=bool, default=True, help='enable CUDA, default as True')
    parser.add_argument('--seed', type=int, default=42, help='set the random seed for stabilizing experiment results')
    parser.add_argument('--dataset', type=str, default='link', choices=['metr-la', 'pems-bay', 'pemsd7-m', 'link'])
    parser.add_argument('--n_pred', type=int, default=1, help='the number of time interval for predcition, default as 1')
    parser.add_argument('--Kt', type=int, default=3, help='Kernel size of temporal block (casual convolution) = (Kt,1)')
    parser.add_argument('--stblock_num', type=int, default=2)
    parser.add_argument('--act_func', type=str, default='glu', choices=['glu', 'gtu'])
    parser.add_argument('--Ks', type=int, default=3, choices=[3, 2], help='Kernel size of Spatial block')
    parser.add_argument('--graph_conv_type', type=str, default='cheb_graph_conv', choices=['cheb_graph_conv', 'graph_conv'])
    parser.add_argument('--gso_type', type=str, default='sym_norm_lap', choices=['sym_norm_lap', 'rw_norm_lap', 'sym_renorm_adj', 'rw_renorm_adj'])
    parser.add_argument('--enable_bias', type=bool, default=True, help='default as True')
    parser.add_argument('--droprate', type=float, default=0.5)
    parser.add_argument('--lr', type=float, default=0.001, help='learning rate')
    parser.add_argument('--weight_decay_rate', type=float, default=0.0005, help='weight decay (L2 penalty)')
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--opt', type=str, default='adam', help='optimizer, default as adam')
    parser.add_argument('--step_size', type=int, default=10, help='period of learning rate decay')
    parser.add_argument('--gamma', type=float, default=0.95)
    parser.add_argument('--patience', type=int, default=30, help='early stopping patience')
    args = parser.parse_args()
    print('Training configs: {}'.format(args))
    logger.info('Training configs: {}'.format(args))

    # For stable experiment results
    set_env(args.seed)

    # Running in Nvidia GPU (CUDA) or CPU
    if args.enable_cuda and torch.cuda.is_available():
        # Set available CUDA devices
        # This option is crucial for multiple GPUs
        # 'cuda' ≡ 'cuda:0'
        device = torch.device('cuda')
        print('running on GPU')
    else:
        device = torch.device('cpu')
        print('running on CPU')
        
    # output layers
    Ko = config.model_config.n_his - (args.Kt - 1) * 2 * args.stblock_num

    # blocks: settings of channel size in st_conv_blocks and output layer,
    # using the bottleneck design in st_conv_blocks
    blocks = []
    blocks.append([1])
    for l in range(args.stblock_num):
        blocks.append([64, 16, 64])
    if Ko == 0:
        blocks.append([128])
    elif Ko > 0:
        blocks.append([128, 128])
    blocks.append([1])
    
    return args, device, blocks

def prepare_model(args, blocks, n_vertex):
    loss = nn.MSELoss()
    es = earlystopping.EarlyStopping(mode='min', min_delta=0.0, patience=args.patience)

    if args.graph_conv_type == 'cheb_graph_conv':
        model = models.STGCNChebGraphConv(args, blocks, n_vertex).to(device)
    else:
        model = models.STGCNGraphConv(args, blocks, n_vertex).to(device)

    if args.opt == "rmsprop":
        optimizer = optim.RMSprop(model.parameters(), lr=args.lr, weight_decay=args.weight_decay_rate)
    elif args.opt == "adam":
        optimizer = optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay_rate, amsgrad=False)
    elif args.opt == "adamw":
        optimizer = optim.AdamW(model.parameters(), lr=args.lr, weight_decay=args.weight_decay_rate, amsgrad=False)
    else:
        raise NotImplementedError(f'ERROR: The optimizer {args.opt} is not implemented.')

    scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=args.step_size, gamma=args.gamma)

    return loss, es, model, optimizer, scheduler


def train(loss, optimizer, scheduler, es, model, train_iter, val_iter, datetime, epochs, save_final=True, save_plot=True):
    best_valid_loss = None
    train_losses, val_losses = [], []
    for epoch in range(epochs):
        l_sum, n = 0.0, 0  # 'l_sum' is epoch sum loss, 'n' is epoch instance number
        model.train()
        for x, y in tqdm.tqdm(train_iter):
            y_pred = model(x).view(len(x), -1)  # [batch_size, num_nodes]
            l = loss(y_pred, y)
            optimizer.zero_grad()
            l.backward()
            optimizer.step()
            scheduler.step()
            l_sum += l.item() * y.shape[0]
            n += y.shape[0]

        train_losses.append(l_sum/n)

        val_loss = val(model, val_iter)
        val_losses.append(val_loss)

        # Save best model
        best_valid_loss, best_epoch = utility.SaveBestModel(save_model_path = TRAINED_MODEL_DIR, 
                                        best_valid_loss = best_valid_loss,current_valid_loss = val_loss, 
                                        epoch=epoch, model=model, datetime=datetime)
        if best_epoch != None:
            logger.info(f"Best validation loss: {best_valid_loss} at epoch {best_epoch}")
        # GPU memory usage
        gpu_mem_alloc = torch.cuda.max_memory_allocated() / 1000000 if torch.cuda.is_available() else 0
        t = torch.cuda.get_device_properties(0).total_memory
        r = torch.cuda.memory_reserved(0)
        a = torch.cuda.memory_allocated(0)
        f = r-a  # free inside reserved
        print('Epoch: {:03d} | Lr: {:.20f} |Train loss: {:.6f} | Val loss: {:.6f} | GPU occupy: {:.6f} MiB'.\
            format(epoch+1, optimizer.param_groups[0]['lr'], l_sum / n, val_loss, gpu_mem_alloc))
        print(f"t:{t/1000000}, r:{r/1000000}, a:{a/1000000}, f:{f/1000000}")
        if es.step(val_loss):
            print('Early stopping.')
            logger.info(f'Early stopping at epoch {epoch}')
            break
    if save_final == True:
        utility.save_model(TRAINED_MODEL_DIR, model, datetime=datetime)
    if save_plot == True:
        utility.save_plots(PLOT_DIR, train_losses, val_losses, datetime)
    
    last_epoch = epoch+1
    logger.info(f'Last epoch: {last_epoch}')
    return model

@torch.no_grad()
def val(model, val_iter):
    model.eval()
    l_sum, n = 0.0, 0
    for x, y in val_iter:
        y_pred = model(x).view(len(x), -1)
        l = loss(y_pred, y)
        l_sum += l.item() * y.shape[0]
        n += y.shape[0]
    return torch.tensor(l_sum / n)

@torch.no_grad() 
def test(zscore, loss, model, test_iter, args):
    model.eval()
    test_MSE = utility.evaluate_model(model, loss, test_iter)
    test_MAE, test_RMSE, test_MAPE, test_WAPE,std_mae, std_rmse, std_mape, max_error, y_true, y_pred = utility.evaluate_metric(model, test_iter, zscore)
    logger.info(f'Dataset {args.dataset:s} | Test loss(MSE) {test_MSE:.6f} | MAE {test_MAE:.6f} std {std_mae:.6f} | RMSE {test_RMSE:.6f} std {std_rmse:.6f} \
| ACC {100-100*test_MAPE:.8f} % | MAPE {100*test_MAPE:.8f} std {100*std_mape:.8f} %| WAPE {100*test_WAPE:.8f} % | MAX_ERROR {max_error:.8f}\n')
    print((f'Dataset {args.dataset:s} | Test loss(MSE) {test_MSE:.6f} | MAE {test_MAE:.6f} std {std_mae:.6f} | RMSE {test_RMSE:.6f} std {std_rmse:.6f} \
| ACC {100-100*test_MAPE:.8f} % | MAPE {100*test_MAPE:.8f} std {100*std_mape:.8f} %| WAPE {100*test_WAPE:.8f} % | MAX_ERROR {max_error:.8f}'))
    return y_true, y_pred

def mean_absolute_percentage_error(y_true, y_pred): 
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    y_true +=  1e-18 #add small values to true velocities to avoid division by zeros
    return np.mean(np.abs((y_true - y_pred) / y_true)) * 100

def plotting(y_true, y_preds, stop_num=0):
    y_sample = y_true[:, stop_num]
    y_pred_sample = y_preds[:, stop_num]

    plt.figure(figsize=(17,10))
    plt.plot(y_sample, label='Ground Truth')
    plt.plot(y_pred_sample, label='Predictions')
    plt.xlabel('time_step')
    plt.ylabel('Link_time(sec)')
    plt.legend(loc="upper left")
    plt.show()



if __name__ == "__main__":
    # Logging
    logger = logging.getLogger('STGCN_TRAIN')
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:%(message)s')
    file_handler = logging.FileHandler('stgcn.log')
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


    args, device, blocks = get_parameters()


    # dataset
    data_path = DATASET_DIR
    n_vertex, zscore, train_iter, val_iter, test_iter, train_data, val_data, test_data = dataloader.data_preparate(args, device, data_path)
    print('Train shape:', train_iter.dataset.tensors[0].shape, train_iter.dataset.tensors[1].shape)
    print("Validation shape:", val_iter.dataset.tensors[0].shape, val_iter.dataset.tensors[1].shape)
    print("Test shape:", test_iter.dataset.tensors[0].shape, test_iter.dataset.tensors[1].shape)
    logger.info(f'Train shape: {train_iter.dataset.tensors[0].shape}, {train_iter.dataset.tensors[1].shape}')
    logger.info(f"Validation shape: {val_iter.dataset.tensors[0].shape}, {val_iter.dataset.tensors[1].shape}")
    logger.info(f"Test shape: {test_iter.dataset.tensors[0].shape}, {test_iter.dataset.tensors[1].shape}")
    # x_test = test_data.tensors[0].cpu()
    # y_test = test_data.tensors[1].cpu()
    # test_data = utils.data.TensorDataset(x_test, y_test)
    torch.save(test_data, os.path.join(TEST_DATA_DIR,'test_data.pt')) # save data
    torch.save(zscore, os.path.join(TEST_DATA_DIR,'zscore.pt')) # save zscore


    # prep model & train
    loss, es, model, optimizer, scheduler = prepare_model(args, blocks, n_vertex)
    datetime = datetime.now().strftime('%d-%m-%y_%H-%M-%S')
    model = train(loss, optimizer, scheduler, es, model, train_iter, val_iter, datetime, config.model_config.epochs)


    # Test
    if config.model_config.epochs > 1:
        model_path = os.path.join(TRAINED_MODEL_DIR,f"best_model_{datetime}.pth")
    else:
        model_path = os.path.join(TRAINED_MODEL_DIR,f"final_model_{datetime}.pth")
    model = torch.load(model_path)
    y_true, y_pred = test(zscore, loss, model, test_iter, args)

    print(y_true[0].shape)
    true = np.concatenate(y_true, axis=0)
    pred = np.concatenate(y_pred, axis=0)
    print("concat pred shape:", pred.shape)
    print("concat true shape:", true.shape)
    print('Training shape:', train_iter.dataset.tensors[0].shape, train_iter.dataset.tensors[1].shape)
    print("Testing shape:", test_iter.dataset.tensors[0].shape, test_iter.dataset.tensors[1].shape)
    if config.model_config.train_plot == True:
        plotting(y_true=true, y_preds=pred, stop_num=config.model_config.plot_stop_num)





import pandas as pd
import numpy as np
import math

def prep_data(data_path, freq = '30min'):
        for path in ['go','back']:
                data = pd.read_csv(data_path).drop(['gps_imei'],axis=1).reset_index(drop=True)
                data = data[(data['path'] == path)]

                link_name_list = data.link_name.tolist()
                line_direction_link_order = []
                for i in range(len(link_name_list)):
                        if len(link_name_list[i]) <= 6 and len(link_name_list[i]) >= 5:
                                line_direction_link_order.append(link_name_list[i][0:2])
                        if len(link_name_list[i]) <= 4:
                                line_direction_link_order.append(link_name_list[i][0:1])
                        if len(link_name_list[i]) >= 7:
                                line_direction_link_order.append(link_name_list[i][0:3])
                data['LineDirectionLinkOrder'] = line_direction_link_order
                data['LineDirectionLinkOrder'] = data['LineDirectionLinkOrder'].astype(int)
        
                data.columns = ['LinkName', 'LinkTravelTime', 'DateTime', 'LineDirectionCode', 'routes', 'LineDirectionLinkOrder']

                # data = data[(0 <= data['LineDirectionLinkOrder']) & (data['LineDirectionLinkOrder'] < limit_stop)]
                # assert len(data['LinkName'].unique()) == limit_stop

                # filter time
                data['DateTime'] = pd.to_datetime(data['DateTime'])
                data['Time_hms'] = data['DateTime'].dt.strftime("%H:%M:%S")
                data = data[data['Time_hms']>= '06:00:00']
                data = data[data['Time_hms']<= '19:00:00']
                data = data.drop('Time_hms', axis=1)

                data.sort_values('DateTime', inplace = True)
                ix = pd.DatetimeIndex(pd.to_datetime(data['DateTime'])).floor(freq)
                data['DateTimeReference'] = pd.to_datetime(ix)
                if path == 'go':
                        all_df = data
                else:
                        all_df = pd.concat([all_df,data],axis=0)

        
        all_df.replace(['go','back'], [0,1], inplace=True)

        cols = ['LineDirectionLinkOrder', 'LineDirectionCode', 'routes']
        all_df['LinkRef'] = all_df[cols].apply(lambda row: '_'.join(row.values.astype(str)), axis=1)

        return all_df.reset_index(drop=True)


def fit_scale(data, freq = '30min', smooth = 1, masking=True):
    means = { }
    scales = { }
    low = { }
    upr = { }
    for k, v in data.sort_values(['routes', 'LineDirectionCode', 'LineDirectionLinkOrder']).groupby('LinkRef', sort = False):
        v_nonan = v[~np.isnan(v['LinkTravelTime'])]

        low[k] = v['LinkTravelTime'].quantile(0.1)
        upr[k] = v['LinkTravelTime'].quantile(0.9)

        ix_ref = pd.DatetimeIndex(pd.to_datetime(v['DateTimeReference']))
        v_ = v.set_index(ix_ref)
        # print(v_[v_.LinkRef=='1_2'].sort_values('DateTime'))

        mean = v_[(low[k] <= v_['LinkTravelTime']) & (v_['LinkTravelTime'] <= upr[k])]['LinkTravelTime'].resample(freq).mean()
        mean = mean.interpolate().rolling(window = smooth, center = True).mean() # linear interpolation
        means[k] = mean 
        #scales[k] = v_[(low[k] < v_['LinkTravelTime']) & (v_['LinkTravelTime'] < upr[k])]['LinkTravelTime'].std()
        scales[k] = 1

    # means_df = pd.DataFrame(data = means)
    means_df = pd.DataFrame(data = means).fillna(method='pad').fillna(method='bfill')

    # Masking
    if masking==True:
        means_df = means_df[means_df.index.hour>=6]
        means_df = means_df[means_df.index.hour<=19]
    means_df = means_df[means_df.columns.drop(list(means_df.filter(regex='_1_35')))]
    return (means_df, scales, low, upr, ix_ref)


def fit_scale_no_inter(data, freq = '30min', fillna=False, masking=True):
    means = { }
    scales = { }
    low = { }
    upr = { }
    for k, v in data.sort_values(['routes', 'LineDirectionCode', 'LineDirectionLinkOrder']).groupby('LinkRef', sort = False):
        v_nonan = v[~np.isnan(v['LinkTravelTime'])]

        low[k] = v['LinkTravelTime'].quantile(0.1)
        upr[k] = v['LinkTravelTime'].quantile(0.9)

        ix_ref = pd.DatetimeIndex(pd.to_datetime(v['DateTimeReference']))
        v_ = v.set_index(ix_ref)
        # print(v_[v_.LinkRef=='1_2'].sort_values('DateTime'))

        mean = v_[(low[k] <= v_['LinkTravelTime']) & (v_['LinkTravelTime'] <= upr[k])]['LinkTravelTime'].resample(freq).mean()
        # mean = mean.interpolate().rolling(window = smooth, center = True).mean() # linear interpolation
        means[k] = mean
        scales[k] = 1

    if fillna==False:
        means_df = pd.DataFrame(data = means)
    else:
        means_df = pd.DataFrame(data = means).fillna(method='pad').fillna(method='bfill')

    # Masking
    if masking==True:
        means_df = means_df[means_df.index.hour>=6]
        means_df = means_df[means_df.index.hour<=19]
    means_df = means_df[means_df.columns.drop(list(means_df.filter(regex='_1_35')))]
    return (means_df, scales, low, upr, ix_ref)

def adj_prep(means):
    adj = []
    all_stop_num = means.columns
    for i in range(len(all_stop_num)-1):
        tmp = np.zeros(len(all_stop_num))
        info_current = all_stop_num[i].split('_')
        info_next = all_stop_num[i+1].split('_')
        if i > 0:
            info_prev = all_stop_num[i-1].split('_')

        tmp[i] = 1
        if info_current[1] == info_next[1] and info_current[2] == info_next[2]:
            tmp[i+1] = 1
            if i > 0:
                tmp[i-1] = 1
        if info_current[1] == info_next[1] and info_current[2] != info_next[2]:
            tmp[i+1] = 0
            if i > 0:
                tmp[i-1] = 1
        if info_current[1] != info_next[1] and info_current[2] == info_next[2]:
            tmp[i+1] = 0
            if i > 0:
                tmp[i-1] = 1
        if i > 0:
            if info_prev[1] != info_current[1] or info_prev[2] != info_current[2]:
                tmp[i-1]=0

        adj.append(tmp)

        if i == len(all_stop_num)-2:
            tmp = np.zeros(len(all_stop_num))
            tmp[i+1] = 1
            adj.append(tmp)
    adj = pd.DataFrame(adj, columns = means.columns)
    return adj

def null_values(means, means_no_inter, p=0):
    null_col = means_no_inter.isnull().sum().sort_values(ascending=False)
    val = math.floor(means_no_inter.shape[0]*(p/100))
    null_len = len(null_col[null_col>val])
    # print(f'Number of stations with null that has null values more than {p}% or {val} rows: {null_len} rows or {math.floor(null_len*100/len(null_col))}%')
    
    null = pd.DataFrame(null_col).reset_index()
    null.rename(columns={0:'num_null'},inplace=True)
    null[['station_num','path','route']] = null['index'].str.split('_',2,expand=True)
    null = null.drop('index', axis=1)
    large_null = null[null.num_null > val]
    group_large_null = large_null.groupby(['route','path']).sum()

    num_cells=[]
    for route in ['6', '56', '133', '35']:
        path_list = ['0','1']
        if route == '35':
            path_list = ['0']
        for path in path_list:
            route_num = path + '_' + route
            col = [i for i in means.columns.tolist() if i.endswith(route_num)]
            # print(col)
            num_cells.append(len(col)*len(means))
    num_cells = pd.DataFrame(num_cells, columns=['num_cells'])
    num_cells['route'] = np.repeat(['6', '56', '133', '35'],2)[:-1]
    num_cells['path'] = np.array(['0','1','0','1','0','1','0'])
    g = group_large_null.reset_index()
    g = g.sort_values('route').reset_index(drop=True)
    g['total_cells'] = num_cells.sort_values('route').reset_index(drop=True)['num_cells']
    g['percent'] = g['num_null']*100/g['total_cells']
    g['percent'] = g['percent'].round(decimals=2)
    return g


import os
import numpy as np
import pandas as pd
import scipy.sparse as sp
import torch
import math
from sklearn import preprocessing
import torch.utils as utils
from regression_model.config.core import *
from regression_model.script import utility



def load_adj(dataset_name):
    dataset_path = './data'
    dataset_path = os.path.join(dataset_path, dataset_name)
    adj = sp.load_npz(os.path.join(dataset_path, 'adj.npz'))
    adj = adj.tocsc()
    
    if dataset_name == 'metr-la':
        n_vertex = 207 # number of nodes
    elif dataset_name == 'pems-bay':
        n_vertex = 325
    elif dataset_name == 'pemsd7-m':
        n_vertex = 228

    return adj, n_vertex

def load_adj_link(dataset_name, dataset_path):
    dataset_path = os.path.join(dataset_path, dataset_name)
    adj = np.load(os.path.join(dataset_path, 'adj.npy'))
    
    if dataset_name == 'link':
        n_vertex = len(adj) # number of nodes

    return adj, n_vertex

def load_data(dataset_name, len_train, len_val, dataset_path):
    dataset_path = os.path.join(dataset_path, dataset_name)
    if dataset_name == 'link':
        vel = pd.read_csv(os.path.join(dataset_path, 'time.csv'))
    else:
        vel = pd.read_csv(os.path.join(dataset_path, 'vel.csv'))

    train = vel[: len_train]
    val = vel[len_train: len_train + len_val]
    test = vel[len_train + len_val:]
    return train, val, test

def data_transform(data, n_his, n_pred, device):
    # produce data slices for x_data and y_data

    n_vertex = data.shape[1]
    len_record = len(data)
    num = len_record - n_his - n_pred
    
    x = np.zeros([num, 1, n_his, n_vertex])
    y = np.zeros([num, n_vertex])
    
    for i in range(num):
        head = i
        tail = i + n_his
        x[i, :, :, :] = data[head: tail].reshape(1, n_his, n_vertex)
        y[i] = data[tail + n_pred - 1]

    return torch.Tensor(x).to(device), torch.Tensor(y).to(device)


def data_preparate(args, device, data_path):
    if args.dataset == 'link':
        adj, n_vertex = load_adj_link(args.dataset, data_path)
        gso = utility.calc_gso(adj, args.gso_type)
        if args.graph_conv_type == 'cheb_graph_conv':
            gso = utility.calc_chebynet_gso(gso)
        gso = gso.toarray()
    else:    
        adj, n_vertex = load_adj(args.dataset)
        gso = utility.calc_gso(adj, args.gso_type)
        if args.graph_conv_type == 'cheb_graph_conv':
            gso = utility.calc_chebynet_gso(gso)
        gso = gso.toarray()

    gso = gso.astype(dtype=np.float32)
    args.gso = torch.from_numpy(gso).to(device)
    print("gso datatype:", gso.dtype)
    print("gps data shape: ", gso.shape)
    dataset_path = os.path.join(data_path, args.dataset)

    if args.dataset == 'link':
        data_col = pd.read_csv(os.path.join(dataset_path, 'time.csv')).shape[0]
    else:
        data_col = pd.read_csv(os.path.join(dataset_path, 'vel.csv')).shape[0]
    # recommended dataset split rate as train: val: test = 60: 20: 20, 70: 15: 15 or 80: 10: 10
    # using dataset split rate as train: val: test = 70: 15: 15
    val_and_test_rate = 0.15

    len_val = int(math.floor(data_col * val_and_test_rate/2))
    len_test = int(math.floor(data_col * val_and_test_rate))
    len_train = int(data_col - len_val - len_test)

    # data_path = r'C:\Users\mizte\Documents\ea\tsb\gnn\deployment\data' # data path
    train, val, test = load_data(args.dataset, len_train, len_val, data_path)
    zscore = preprocessing.StandardScaler()
    train = zscore.fit_transform(train)
    val = zscore.transform(val)
    test = zscore.transform(test)

    x_train, y_train = data_transform(train, config.model_config.n_his, args.n_pred, device)
    x_val, y_val = data_transform(val, config.model_config.n_his, args.n_pred, device)
    x_test, y_test = data_transform(test, config.model_config.n_his, args.n_pred, device)
    train_data = utils.data.TensorDataset(x_train, y_train)
    train_iter = utils.data.DataLoader(dataset=train_data, batch_size=args.batch_size, shuffle=False)
    val_data = utils.data.TensorDataset(x_val, y_val)
    val_iter = utils.data.DataLoader(dataset=val_data, batch_size=args.batch_size, shuffle=False)
    test_data = utils.data.TensorDataset(x_test, y_test)
    test_iter = utils.data.DataLoader(dataset=test_data, batch_size=args.batch_size, shuffle=False)

    return n_vertex, zscore, train_iter, val_iter, test_iter, train_data, val_data, test_data


from typing import Dict, List, Sequence

from pathlib import Path
from pydantic import BaseModel
from strictyaml import YAML, load

import regression_model

# Directory
PACKAGE_ROOT = Path(regression_model.__file__).resolve().parent
ROOT = PACKAGE_ROOT.parent
CONFIG_FILE_PATH = PACKAGE_ROOT / "config.yml"
DATASET_DIR = PACKAGE_ROOT / "data"
TRAINED_MODEL_DIR = PACKAGE_ROOT / "save_model"
TEST_DATA_DIR = PACKAGE_ROOT / 'train_val_test'
PLOT_DIR = PACKAGE_ROOT / 'plots'


class ModelConfig(BaseModel):
    package_name: str
    epochs: int
    plot_stop_num: int
    train_plot: bool
    test_size: float
    n_his: int
    test_model: str
    test_plot: bool

class Config(BaseModel):
    model_config: ModelConfig


def find_config_file() -> Path:
    if CONFIG_FILE_PATH.is_file():
        return CONFIG_FILE_PATH
    raise Exception(f'Config not found at {CONFIG_FILE_PATH}')


def fetch_config_from_yaml(cfg_path: Path = None) -> YAML:
    """Parse YAML containing the package configuration."""

    if not cfg_path:
        cfg_path = find_config_file()

    if cfg_path:
        with open(cfg_path, "r") as conf_file:
            parsed_config = load(conf_file.read())
            return parsed_config
    raise OSError(f"Did not find config file at path: {cfg_path}")


def validation_config(parsed_config: YAML = None) -> Config:
    """Run validation on config values."""
    if parsed_config is None:
        parsed_config = fetch_config_from_yaml()

    # specify the data attribute from the strictyaml YAML type.
    _config = Config(
        model_config=ModelConfig(**parsed_config.data),
    )

    return _config

config = validation_config()
print("PACKAGE_ROOT: ", PACKAGE_ROOT)


import torch
from regression_model.config.core import *
from regression_model.script import utility


@torch.no_grad() 
def make_pred(zscore, test_model, test_iter):
    test_model.eval()
    test_MAE, test_RMSE, test_MAPE, test_WAPE,std_mae, std_rmse, std_mape, max_error, y_true, y_pred = utility.evaluate_metric(test_model, test_iter, zscore)
    print((f'Dataset Link | MAE {test_MAE:.6f} std {std_mae:.6f} | RMSE {test_RMSE:.6f} std {std_rmse:.6f} \
| ACC {100-100*test_MAPE:.8f} % | MAPE {100*test_MAPE:.8f} std {100*std_mape:.8f} %| WAPE {100*test_WAPE:.8f} %'))
    return y_true, y_pred


def test_pred(sample_input_data):
    test_iter, zscore, model = sample_input_data    
    assert test_iter.dataset.tensors[1].shape[1] == 491

    y_true, y_pred = make_pred(zscore, model, test_iter)
    assert len(y_true) == len(y_pred)


import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import eigsh
import matplotlib.pyplot as plt
import os
import torch
from regression_model.config.core import *


plt.style.use('ggplot')

def calc_gso(dir_adj, gso_type):
    n_vertex = dir_adj.shape[0]

    if sp.issparse(dir_adj) == False:
        dir_adj = sp.csc_matrix(dir_adj)
    elif dir_adj.format != 'csc':
        dir_adj = dir_adj.tocsc()

    id = sp.identity(n_vertex, format='csc') # identity matrix

    # Symmetrizing an adjacency matrix
    adj = dir_adj + dir_adj.T.multiply(dir_adj.T > dir_adj) - dir_adj.multiply(dir_adj.T > dir_adj)
    #adj = 0.5 * (dir_adj + dir_adj.transpose())
    
    if gso_type == 'sym_renorm_adj' or gso_type == 'rw_renorm_adj' \
        or gso_type == 'sym_renorm_lap' or gso_type == 'rw_renorm_lap':
        adj = adj + id
    
    if gso_type == 'sym_norm_adj' or gso_type == 'sym_renorm_adj' \
        or gso_type == 'sym_norm_lap' or gso_type == 'sym_renorm_lap':
        row_sum = adj.sum(axis=1).A1
        row_sum_inv_sqrt = np.power(row_sum, -0.5)
        row_sum_inv_sqrt[np.isinf(row_sum_inv_sqrt)] = 0.
        deg_inv_sqrt = sp.diags(row_sum_inv_sqrt, format='csc')
        # A_{sym} = D^{-0.5} * A * D^{-0.5}
        sym_norm_adj = deg_inv_sqrt.dot(adj).dot(deg_inv_sqrt)

        if gso_type == 'sym_norm_lap' or gso_type == 'sym_renorm_lap':
            sym_norm_lap = id - sym_norm_adj
            gso = sym_norm_lap
        else:
            gso = sym_norm_adj

    elif gso_type == 'rw_norm_adj' or gso_type == 'rw_renorm_adj' \
        or gso_type == 'rw_norm_lap' or gso_type == 'rw_renorm_lap':
        row_sum = np.sum(adj, axis=1).A1
        row_sum_inv = np.power(row_sum, -1)
        row_sum_inv[np.isinf(row_sum_inv)] = 0.
        deg_inv = np.diag(row_sum_inv)
        # A_{rw} = D^{-1} * A
        rw_norm_adj = deg_inv.dot(adj)

        if gso_type == 'rw_norm_lap' or gso_type == 'rw_renorm_lap':
            rw_norm_lap = id - rw_norm_adj
            gso = rw_norm_lap
        else:
            gso = rw_norm_adj

    else:
        raise ValueError(f'{gso_type} is not defined.')

    return gso

def calc_chebynet_gso(gso):
    if sp.issparse(gso) == False:
        gso = sp.csc_matrix(gso)
    elif gso.format != 'csc':
        gso = gso.tocsc()

    id = sp.identity(gso.shape[0], format='csc')
    eigval_max = max(eigsh(A=gso, k=6, which='LM', return_eigenvectors=False))

    # If the gso is symmetric or random walk normalized Laplacian,
    # then the maximum eigenvalue is smaller than or equals to 2.
    if eigval_max >= 2:
        gso = gso - id
    else:
        gso = 2 * gso / eigval_max - id

    return gso


def cnv_sparse_mat_to_coo_tensor(sp_mat, device):
    # convert a compressed sparse row (csr) or compressed sparse column (csc) matrix to a hybrid sparse coo tensor
    sp_coo_mat = sp_mat.tocoo()
    i = torch.from_numpy(np.vstack((sp_coo_mat.row, sp_coo_mat.col)))
    v = torch.from_numpy(sp_coo_mat.data)
    s = torch.Size(sp_coo_mat.shape)

    if sp_mat.dtype == np.float32 or sp_mat.dtype == np.float64:
        return torch.sparse_coo_tensor(indices=i, values=v, size=s, dtype=torch.float32, device=device, requires_grad=False)
    else:
        raise TypeError(f'ERROR: The dtype of {sp_mat} is {sp_mat.dtype}, not been applied in implemented models.')

def evaluate_model(model, loss, data_iter):
    model.eval()
    l_sum, n = 0.0, 0
    with torch.no_grad():
        for x, y in data_iter:
            y_pred = model(x).view(len(x), -1)
            l = loss(y_pred, y)
            l_sum += l.item() * y.shape[0]
            n += y.shape[0]
        mse = l_sum / n
    return mse

def evaluate_metric(model, data_iter, scaler):
    y_true = list()
    y_predictions = list()
    model.eval()
    with torch.no_grad():
        mae, sum_y, mape, mse = [], [], [], []
        r2 = []
        for x, y in data_iter:
            y_scaled = scaler.inverse_transform(y.cpu().numpy())
            y_pred_scaled = scaler.inverse_transform(model(x).view(len(x), -1).cpu().numpy())

            y_true.append(y_scaled)
            y_predictions.append(y_pred_scaled)

            y = y_scaled.reshape(-1)
            y_pred = y_pred_scaled.reshape(-1)
            d = np.abs(y - y_pred)
            mae += d.tolist()
            sum_y += y.tolist()
            mape += (d / y).tolist()
            mse += (d ** 2).tolist()
        MAE = np.array(mae).mean()
        MAPE = np.array(mape).mean()
        RMSE = np.sqrt(np.array(mse).mean())
        WAPE = np.sum(np.array(mae)) / np.sum(np.array(sum_y))
        std_mae = np.std(np.array(mae))
        std_rmse = np.std(np.sqrt((np.array(mse))))
        std_mape = np.std(np.array(mape))
        max_error = np.max(mae)
        # max_argmax = np.argmax(mae)

        #return MAE, MAPE, RMSE
        return MAE, RMSE, MAPE, WAPE,std_mae, std_rmse, std_mape, max_error, y_true, y_predictions

def load_model(model_path = 'best_model.pth', zscore_path = 'zscore.pt'):
    model = torch.load(os.path.join(TRAINED_MODEL_DIR, model_path))
    scaler = torch.load(os.path.join(TEST_DATA_DIR, zscore_path))
    return model, scaler

def make_prediction(model, scaler, one_pred_x):
    model.eval()
    with torch.no_grad():
        pred = model(one_pred_x).view(len(one_pred_x), -1).cpu().numpy()
        scaled_pred = scaler.inverse_transform(pred)
    return scaled_pred

def SaveBestModel(save_model_path, best_valid_loss, current_valid_loss, epoch, model, datetime):
    if epoch == 0:
        best_valid_loss = current_valid_loss

    if current_valid_loss < best_valid_loss:
        best_valid_loss = current_valid_loss
        print(f"\nBest validation loss: {best_valid_loss} at epoch {epoch}")
        torch.save(model, os.path.join(save_model_path,f'best_model_{datetime}.pth'))
        return best_valid_loss, epoch

    return best_valid_loss, None

def save_model(save_model_path, model, datetime):
    """
    Function to save the trained model to disk.
    """
    print(f"Saving final model...")
    torch.save(model, os.path.join(save_model_path,f'final_model_{datetime}.pth'))

def save_plots(plot_path, train_loss, valid_loss, datetime):
    """
    Function to save the loss and accuracy plots to disk.
    """ 
    # loss plots
    plt.figure(figsize=(10, 7))
    plt.plot(
        train_loss, color='orange', linestyle='-', 
        label='train loss'
    )
    plt.plot(
        valid_loss, color='red', linestyle='-', 
        label='validataion loss'
    )
    plt.xlabel('Epochs')
    plt.ylabel('Loss(MSE)')
    plt.legend()
    plt.savefig(os.path.join(plot_path, f'loss_{datetime}.png'))





import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

class Align(nn.Module):
    def __init__(self, c_in, c_out):
        super(Align, self).__init__()
        self.c_in = c_in
        self.c_out = c_out
        self.align_conv = nn.Conv2d(in_channels=c_in, out_channels=c_out, kernel_size=(1, 1))

    def forward(self, x):
        if self.c_in > self.c_out:
            x = self.align_conv(x)
        elif self.c_in < self.c_out:
            batch_size, _, timestep, n_vertex = x.shape
            x = torch.cat([x, torch.zeros([batch_size, self.c_out - self.c_in, timestep, n_vertex]).to(x)], dim=1)
        else:
            x = x
        
        return x

class CausalConv1d(nn.Conv1d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, enable_padding=False, dilation=1, groups=1, bias=True):
        if enable_padding == True:
            self.__padding = (kernel_size - 1) * dilation
        else:
            self.__padding = 0
        super(CausalConv1d, self).__init__(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=self.__padding, dilation=dilation, groups=groups, bias=bias)

    def forward(self, input):
        result = super(CausalConv1d, self).forward(input)
        if self.__padding != 0:
            return result[: , : , : -self.__padding] # Chomp1d
        
        return result

class CausalConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, enable_padding=False, dilation=1, groups=1, bias=True):
        kernel_size = nn.modules.utils._pair(kernel_size)
        stride = nn.modules.utils._pair(stride)
        dilation = nn.modules.utils._pair(dilation) # conventional convolutional layer is 1-dilated layer
        if enable_padding == True:
            self.__padding = [int((kernel_size[i] - 1) * dilation[i]) for i in range(len(kernel_size))]
        else:
            self.__padding = 0
        self.left_padding = nn.modules.utils._pair(self.__padding)
        super(CausalConv2d, self).__init__(in_channels, out_channels, kernel_size, stride=stride, padding=0, dilation=dilation, groups=groups, bias=bias)
        
    def forward(self, input):
        if self.__padding != 0:
            input = F.pad(input, (self.left_padding[1], 0, self.left_padding[0], 0)) # (left_pad, right_pad, top_pad, bottom_pad)
        result = super(CausalConv2d, self).forward(input)

        return result

class TemporalConvLayer(nn.Module):

    # Temporal Convolution Layer (GLU)
    #
    #        |--------------------------------| * Residual Connection *
    #        |                                |
    #        |    |--->--- CasualConv2d ----- + -------|       
    # -------|----|                                   ⊙ ------>
    #             |--->--- CasualConv2d --- Sigmoid ---|                               
    #
    
    #param x: tensor, [bs, c_in, ts, n_vertex]

    def __init__(self, Kt, c_in, c_out, n_vertex, act_func):
        super(TemporalConvLayer, self).__init__()
        self.Kt = Kt
        self.c_in = c_in # Channel input
        self.c_out = c_out # Channel output
        self.n_vertex = n_vertex
        self.align = Align(c_in, c_out)
        if act_func == 'glu' or act_func == 'gtu':
            self.causal_conv = CausalConv2d(in_channels=c_in, out_channels=2 * c_out, kernel_size=(Kt, 1), enable_padding=False, dilation=1)
        else:
            self.causal_conv = CausalConv2d(in_channels=c_in, out_channels=c_out, kernel_size=(Kt, 1), enable_padding=False, dilation=1)
        self.act_func = act_func
        self.sigmoid = nn.Sigmoid()
        self.tanh = nn.Tanh()
        self.relu = nn.ReLU()
        self.leaky_relu = nn.LeakyReLU()
        self.silu = nn.SiLU()

    def forward(self, x):   
        x_in = self.align(x)[:, :, self.Kt - 1:, :]
        x_causal_conv = self.causal_conv(x)

        if self.act_func == 'glu' or self.act_func == 'gtu':
            x_p = x_causal_conv[:, : self.c_out, :, :]
            x_q = x_causal_conv[:, -self.c_out:, :, :]

            if self.act_func == 'glu':
                # GLU was first purposed in
                # *Language Modeling with Gated Convolutional Networks*.
                # URL: https://arxiv.org/abs/1612.08083
                # Input tensor X is split by a certain dimension into tensor X_a and X_b.
                # In PyTorch, GLU is defined as X_a ⊙ Sigmoid(X_b).
                # URL: https://pytorch.org/docs/master/nn.functional.html#torch.nn.functional.glu
                # (x_p + x_in) ⊙ Sigmoid(x_q)
                x = torch.mul((x_p + x_in), self.sigmoid(x_q))

            else:
                # Tanh(x_p + x_in) ⊙ Sigmoid(x_q)
                x = torch.mul(self.tanh(x_p + x_in), self.sigmoid(x_q))

        elif self.act_func == 'relu':
            x = self.relu(x_causal_conv + x_in)
        
        elif self.act_func == 'leaky_relu':
            x = self.leaky_relu(x_causal_conv + x_in)

        elif self.act_func == 'silu':
            x = self.silu(x_causal_conv + x_in)
        
        else:
            raise NotImplementedError(f'ERROR: The activation function {self.act_func} is not implemented.')
        
        return x

class ChebGraphConv(nn.Module):
    def __init__(self, c_in, c_out, Ks, gso, bias):
        super(ChebGraphConv, self).__init__()
        self.c_in = c_in
        self.c_out = c_out
        self.Ks = Ks
        self.gso = gso
        self.weight = nn.Parameter(torch.FloatTensor(Ks, c_in, c_out))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(c_out))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)
    
    def forward(self, x):
        #bs, c_in, ts, n_vertex = x.shape
        x = torch.permute(x, (0, 2, 3, 1))

        if self.Ks - 1 < 0:
            raise ValueError(f'ERROR: the graph convolution kernel size Ks has to be a positive integer, but received {self.Ks}.')  
        elif self.Ks - 1 == 0:
            x_0 = x
            x_list = [x_0]
        elif self.Ks - 1 == 1:
            x_0 = x
            x_1 = torch.einsum('hi,btij->bthj', self.gso, x)
            x_list = [x_0, x_1]
        elif self.Ks - 1 >= 2:
            x_0 = x
            x_1 = torch.einsum('hi,btij->bthj', self.gso, x)
            x_list = [x_0, x_1]
            for k in range(2, self.Ks):
                x_list.append(torch.einsum('hi,btij->bthj', 2 * self.gso, x_list[k - 1]) - x_list[k - 2])
        
        x = torch.stack(x_list, dim=2)

        cheb_graph_conv = torch.einsum('btkhi,kij->bthj', x, self.weight)

        if self.bias is not None:
            cheb_graph_conv = torch.add(cheb_graph_conv, self.bias)
        else:
            cheb_graph_conv = cheb_graph_conv
        
        return cheb_graph_conv

class GraphConv(nn.Module):
    def __init__(self, c_in, c_out, gso, bias):
        super(GraphConv, self).__init__()
        self.c_in = c_in
        self.c_out = c_out
        self.gso = gso
        self.weight = nn.Parameter(torch.FloatTensor(c_in, c_out))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(c_out))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        #bs, c_in, ts, n_vertex = x.shape
        x = torch.permute(x, (0, 2, 3, 1))

        first_mul = torch.einsum('hi,btij->bthj', self.gso, x)
        second_mul = torch.einsum('bthi,ij->bthj', first_mul, self.weight)

        if self.bias is not None:
            graph_conv = torch.add(second_mul, self.bias)
        else:
            graph_conv = second_mul
        
        return graph_conv

class GraphConvLayer(nn.Module):
    def __init__(self, graph_conv_type, c_in, c_out, Ks, gso, bias):
        super(GraphConvLayer, self).__init__()
        self.graph_conv_type = graph_conv_type
        self.c_in = c_in
        self.c_out = c_out
        self.align = Align(c_in, c_out)
        self.Ks = Ks
        self.gso = gso
        if self.graph_conv_type == 'cheb_graph_conv':
            self.cheb_graph_conv = ChebGraphConv(c_out, c_out, Ks, gso, bias)
        elif self.graph_conv_type == 'graph_conv':
            self.graph_conv = GraphConv(c_out, c_out, gso, bias)

    def forward(self, x):
        x_gc_in = self.align(x)
        if self.graph_conv_type == 'cheb_graph_conv':
            x_gc = self.cheb_graph_conv(x_gc_in)
        elif self.graph_conv_type == 'graph_conv':
            x_gc = self.graph_conv(x_gc_in)
        x_gc = x_gc.permute(0, 3, 1, 2)
        x_gc_out = torch.add(x_gc, x_gc_in)

        return x_gc_out

class STConvBlock(nn.Module):
    # STConv Block contains 'TGTND' structure
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # G: Graph Convolution Layer (ChebGraphConv or GraphConv)
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # D: Dropout

    def __init__(self, Kt, Ks, n_vertex, last_block_channel, channels, act_func, graph_conv_type, gso, bias, droprate):
        super(STConvBlock, self).__init__()
        self.tmp_conv1 = TemporalConvLayer(Kt, last_block_channel, channels[0], n_vertex, act_func)
        self.graph_conv = GraphConvLayer(graph_conv_type, channels[0], channels[1], Ks, gso, bias)
        self.tmp_conv2 = TemporalConvLayer(Kt, channels[1], channels[2], n_vertex, act_func)
        self.tc2_ln = nn.LayerNorm([n_vertex, channels[2]])
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=droprate)

    def forward(self, x):
        x = self.tmp_conv1(x)
        x = self.graph_conv(x)
        x = self.relu(x)
        x = self.tmp_conv2(x)
        x = self.tc2_ln(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        x = self.dropout(x)

        return x

class OutputBlock(nn.Module):
    # Output block contains 'TNFF' structure
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # F: Fully-Connected Layer
    # F: Fully-Connected Layer

    def __init__(self, Ko, last_block_channel, channels, end_channel, n_vertex, act_func, bias, droprate):
        super(OutputBlock, self).__init__()
        self.tmp_conv1 = TemporalConvLayer(Ko, last_block_channel, channels[0], n_vertex, act_func)
        self.fc1 = nn.Linear(in_features=channels[0], out_features=channels[1], bias=bias)
        self.fc2 = nn.Linear(in_features=channels[1], out_features=end_channel, bias=bias)
        self.tc1_ln = nn.LayerNorm([n_vertex, channels[0]])
        self.relu = nn.ReLU()
        self.leaky_relu = nn.LeakyReLU()
        self.silu = nn.SiLU()
        self.dropout = nn.Dropout(p=droprate)

    def forward(self, x):
        x = self.tmp_conv1(x)
        x = self.tc1_ln(x.permute(0, 2, 3, 1))
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x).permute(0, 3, 1, 2)

        return x



import torch
import torch.nn as nn
from regression_model.config.core import config
from regression_model.model import layers

class STGCNChebGraphConv(nn.Module):
    # STGCNChebGraphConv contains 'TGTND TGTND TNFF' structure
    # ChebGraphConv is the graph convolution from ChebyNet.
    # Using the Chebyshev polynomials of the first kind as a graph filter.
        
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # G: Graph Convolution Layer (ChebGraphConv)
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # D: Dropout

    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # G: Graph Convolution Layer (ChebGraphConv)
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # D: Dropout

    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normalization
    # F: Fully-Connected Layer
    # F: Fully-Connected Layer

    def __init__(self, args, blocks, n_vertex):
        super(STGCNChebGraphConv, self).__init__()
        modules = []
        for l in range(len(blocks) - 3):
            modules.append(layers.STConvBlock(args.Kt, args.Ks, n_vertex, blocks[l][-1], blocks[l+1], args.act_func, args.graph_conv_type, args.gso, args.enable_bias, args.droprate))
        self.st_blocks = nn.Sequential(*modules)
        Ko = config.model_config.n_his - (len(blocks) - 3) * 2 * (args.Kt - 1)
        self.Ko = Ko
        if self.Ko > 1:
            self.output = layers.OutputBlock(Ko, blocks[-3][-1], blocks[-2], blocks[-1][0], n_vertex, args.act_func, args.enable_bias, args.droprate)
        elif self.Ko == 0:
            self.fc1 = nn.Linear(in_features=blocks[-3][-1], out_features=blocks[-2][0], bias=args.enable_bias)
            self.fc2 = nn.Linear(in_features=blocks[-2][0], out_features=blocks[-1][0], bias=args.enable_bias)
            self.relu = nn.ReLU()
            self.leaky_relu = nn.LeakyReLU()
            self.silu = nn.SiLU()
            self.dropout = nn.Dropout(p=args.droprate)

    def forward(self, x):
        x = self.st_blocks(x)
        if self.Ko > 1:
            x = self.output(x)
        elif self.Ko == 0:
            x = self.fc1(x.permute(0, 2, 3, 1))
            x = self.relu(x)
            x = self.fc2(x).permute(0, 3, 1, 2)
        
        return x

class STGCNGraphConv(nn.Module):
    # STGCNGraphConv contains 'TGTND TGTND TNFF' structure
    # GraphConv is the graph convolution from GCN.
    # GraphConv is not the first-order ChebConv, because the renormalization trick is adopted.
    # Be careful about over-smoothing.
    
    # Architecture
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # G: Graph Convolution Layer (GraphConv)
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # D: Dropout

    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # G: Graph Convolution Layer (GraphConv)
    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normolization
    # D: Dropout

    # T: Gated Temporal Convolution Layer (GLU or GTU)
    # N: Layer Normalization
    # F: Fully-Connected Layer
    # F: Fully-Connected Layer

    def __init__(self, args, blocks, n_vertex):
        super(STGCNGraphConv, self).__init__()
        modules = []
        for l in range(len(blocks) - 3):
            modules.append(layers.STConvBlock(args.Kt, args.Ks, n_vertex, blocks[l][-1], blocks[l+1], args.act_func, args.graph_conv_type, args.gso, args.enable_bias, args.droprate))
        self.st_blocks = nn.Sequential(*modules)
        Ko = config.model_config.n_his - (len(blocks) - 3) * 2 * (args.Kt - 1)
        self.Ko = Ko
        if self.Ko > 1:
            self.output = layers.OutputBlock(Ko, blocks[-3][-1], blocks[-2], blocks[-1][0], n_vertex, args.act_func, args.enable_bias, args.droprate)
        elif self.Ko == 0:
            self.fc1 = nn.Linear(in_features=blocks[-3][-1], out_features=blocks[-2][0], bias=args.enable_bias)
            self.fc2 = nn.Linear(in_features=blocks[-2][0], out_features=blocks[-1][0], bias=args.enable_bias)
            self.relu = nn.ReLU()
            self.leaky_relu = nn.LeakyReLU()
            self.silu = nn.SiLU()
            self.do = nn.Dropout(p=args.droprate)

    def forward(self, x):
        x = self.st_blocks(x)
        if self.Ko > 1:
            x = self.output(x)
        elif self.Ko == 0:
            x = self.fc1(x.permute(0, 2, 3, 1))
            x = self.relu(x)
            x = self.fc2(x).permute(0, 3, 1, 2)
        
        return x


import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils as utils
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.sampler import RandomSampler

import os
import random
import numpy as np
import argparse
import matplotlib.pyplot as plt
from script import utility
import pickle
import matplotlib.lines as mlines
from pathlib import Path
import pandas as pd
from regression_model.config.core import *



def set_env(seed):
    # Set available CUDA devices
    # This option is crucial for an multi-GPU device
    os.environ['CUDA_VISIBLE_DEVICES'] = '0, 1'
    # os.environ['CUDA_LAUNCH_BLOCKING'] = '1'
    # os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':16:8'
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
    os.environ['PYTHONHASHSEED']=str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms(True)

def get_parameters():
    parser = argparse.ArgumentParser(description='STGCN')
    parser.add_argument('--enable_cuda', type=bool, default=True, help='enable CUDA, default as True')
    parser.add_argument('--seed', type=int, default=42, help='set the random seed for stabilizing experiment results')
    parser.add_argument('--dataset', type=str, default='link', choices=['metr-la', 'link'])
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--plot_stop', type=str, default='0_0_6')
    parser.add_argument('--error_route', type=str, default='6')
    args = parser.parse_args()
    # print('Test configs: {}'.format(args))

    # For stable experiment results
    set_env(args.seed)
    
    return args

@torch.no_grad() 
def test(zscore, loss, model, test_iter):
    model.eval()
    test_MSE = utility.evaluate_model(model, loss, test_iter)
    test_MAE, test_RMSE, test_MAPE, test_WAPE,std_mae, std_rmse, std_mape, max_error, y_true, y_pred = utility.evaluate_metric(model, test_iter, zscore)
    print((f'Test loss(MSE) {test_MSE:.6f} | MAE {test_MAE:.6f} std {std_mae:.6f} | RMSE {test_RMSE:.6f} std {std_rmse:.6f} \
| ACC {100-100*test_MAPE:.8f} % | MAPE {100*test_MAPE:.8f} std {100*std_mape:.8f} %| WAPE {100*test_WAPE:.8f} %'))
    return y_true, y_pred

def plotting_index(y_true, y_preds, time_index, stop, stations, show_plot = True):
    with open(stations, 'rb') as f:
        stations = pickle.load(f)
    stop_num = stations.index(stop)

    y_sample = y_true[:, stop_num]
    y_pred_sample = y_preds[:, stop_num]

    with open(time_index, "rb") as f:
        index = pickle.load(f)
    index = index[-len(y_true)-1:-1]

    plt.figure(figsize=(17,10))
    plt.plot(index, y_sample, 'o',label='Ground Truth', color='green')
    plt.plot(index, y_sample,label='Ground Truth', color='green')
    plt.plot(index, y_pred_sample, 'o',label='Predictions', color='red')
    plt.plot(index, y_pred_sample, label='Predictions', color='red')
    plt.xlabel('time')
    plt.ylabel('Link_time(sec)')
    plt.legend(loc="upper left")
    plt.title(f'station: {stop}')
    if show_plot:
        plt.show()
    return index

def plotting(y_true, y_preds, stop, stations):
    with open(stations, 'rb') as f:
        stations = pickle.load(f)
    stop_num = stations.index(stop)

    y_sample = y_true[:, stop_num]
    y_pred_sample = y_preds[:, stop_num]

    plt.figure(figsize=(17,10))
    plt.plot(y_sample,label='Ground Truth', color='green')
    plt.plot(y_pred_sample, label='Predictions', color='red')
    plt.xlabel('time_step')
    plt.ylabel('Link_time(sec)')
    plt.legend(loc="upper left")
    plt.title(f'station: {stop}')
    plt.show()

def show_error(true, pred, time_index):

    with open(time_index, "rb") as f:
        idx = pickle.load(f)
    idx = idx[-len(true)-1:-1]

    with open(os.path.join(DATASET_DIR, 'link', 'stations'), 'rb') as f:
        stations = pickle.load(f)
    err = []
    error_route_list = ['6', '56', '133', '35']
    for error_route in error_route_list:
        path_list = ['0','1']
        if error_route == '35':
            path_list = ['0']
        MAE, ACC, RMSE, MAPE, max_error, max_row, max_col, max_error_true, max_error_pred = [], [], [], [], [], [], [], [], []
        for path in path_list:
            route = path+ '_' + error_route
            station_result = [i for i in stations if i.endswith(route)]
            col = [i for i,v in enumerate(stations) if v in station_result]
            y_true = true[:, col]
            y_pred = pred[:, col]
            d = np.abs(y_true-y_pred)
            mae = d.mean()
            mape = np.abs((y_true-y_pred)/y_true).mean()
            mse = d**2
            ind = np.unravel_index(np.argmax(d, axis=None), d.shape)
            max_row.append(idx[ind[0]])
            max_col.append(station_result[ind[1]])
            max_error.append(d[ind])
            max_error_true.append(y_true[ind])
            max_error_pred.append(y_pred[ind])
            MAE.append(mae)
            MAPE.append(mape*100)
            RMSE.append(np.sqrt(mse.mean()))
            ACC.append((1-mape)*100)

        err.append(pd.DataFrame([MAE, RMSE, MAPE, ACC, max_row, max_col, max_error, max_error_true, max_error_pred]))
    error = pd.concat(err,axis=1).T
    metrics = ['MAE', 'RMSE','MAPE', 'ACC', 'max_error_time', 'max_error_link', 'max_abs_error', 'max_error_true', 'max_error_pred']
    error.columns = metrics
    error.reset_index(inplace=True)
    error.rename(columns={'index':'path'}, inplace=True)
    error.index = np.repeat(error_route_list,2)[:-1]
    error.reset_index(inplace=True)
    error.rename(columns={'index':'route'}, inplace=True)
    error = error.sort_values('route').reset_index(drop=True)
    round_col = ['MAE', 'max_abs_error', 'RMSE', 'MAPE', 'ACC', 'max_error_pred']
    error[round_col] = error[round_col].astype(float).apply(lambda x: pd.Series.round(x,2))
    return error, stations

def confidence(true, pred, error_metric='rmse', test_plot=True):
    diff = np.abs(pred.T - true.T)
    error = np.abs(((true.T-pred.T)/true.T)).tolist()
    mae_nn = np.mean(np.array(diff))
    rmse_nn = np.sqrt((np.array(diff)**2).mean())
    mape_nn = np.mean(np.array(error))
    total_error_list = [mae_nn, rmse_nn, mape_nn*100, (1-mape_nn)*100]
    total_error = pd.DataFrame(total_error_list, columns=['total'])
    total_error.index = ['MAE','RMSE', 'MAPE', 'ACC']

    if error_metric =='mae':
        e_nn = np.mean(diff, axis=1) # per link
    else:
        e_nn = np.sqrt(np.mean((diff**2),axis=1)) # per link

    _, ax = plt.subplots()
    ax.violinplot(
        list(e_nn), showmeans=True, showmedians=False, showextrema=False, widths=1.0
    )

    ax.set_xlabel("Scaled distribution amplitude")
    ax.set_ylabel(f'{error_metric}')
    ax.set_title("Distribution over segments: NN pred")
    # plt.legend(handles=(line1, line2), title="Prediction Model", loc=2)
    plt.savefig(os.path.join(DATASET_DIR, 'link', 'inference_out', 'distributed_error_plot.png'))
    if test_plot == True:
        plt.show()
    return total_error
    


if __name__ == "__main__":
    args = get_parameters()

    # Input
    test_data = torch.load(os.path.join(TEST_DATA_DIR, 'test_data.pt'))
    test_iter = utils.data.DataLoader(dataset=test_data, batch_size=args.batch_size, shuffle=False)
    print("Test data shape:", test_iter.dataset.tensors[0].shape, test_iter.dataset.tensors[1].shape)


    # Output path
    true_path = Path(os.path.join(DATASET_DIR, 'link','true.npy'))
    pred_path = Path(os.path.join(DATASET_DIR, 'link','pred.npy'))


    # Prediction
    loss = nn.MSELoss()
    zscore = torch.load(os.path.join(TEST_DATA_DIR, 'zscore.pt'))
    model = torch.load(os.path.join(TRAINED_MODEL_DIR, config.model_config.test_model))
    y_true, y_pred = test(zscore, loss, model, test_iter)
    true = np.concatenate(y_true, axis=0)
    pred = np.concatenate(y_pred, axis=0)
    np.save(true_path, true)
    np.save(pred_path, pred)


    # Show errors per route
    null = pd.read_csv(os.path.join(DATASET_DIR, 'link','null.csv'))
    error, stations = show_error(true,pred, os.path.join(DATASET_DIR, 'link', 'time_index'))
    print(null)
    print(error)
    error.to_csv(os.path.join(DATASET_DIR, 'link', 'inference_out', 'error_metrics_path_lv.csv'), index=False)
    

    # plotting
    index=[]
    if config.model_config.test_plot == True:
        index = plotting_index(y_true=true, y_preds=pred, time_index=os.path.join(DATASET_DIR, 'link', 'time_index'), 
                                stop=args.plot_stop, stations=os.path.join(DATASET_DIR, 'link', 'stations'), show_plot=False)
    total_error = confidence(true, pred, error_metric='rmse', test_plot=config.model_config.test_plot)
    total_error.to_csv(os.path.join(DATASET_DIR, 'link', 'inference_out', 'total_error.csv'))
    print(total_error)
    if len(index)>0:
        prediction = pd.DataFrame(pred)
        prediction.index=index
        prediction.columns = stations
        prediction.to_csv(os.path.join(DATASET_DIR, 'link', 'inference_out', 'pred_result.csv'), index=True, header=True)